@@ -1,4 +1,6 @@
 {-# LANGUAGE TupleSections #-}
+-- incomplete patterns in 'from{Asc,Desc}List'
+{-# OPTIONS_GHC -Wno-incomplete-uni-patterns #-}
 
 -- | A structure containing unique elements
 module Mini.Data.Set (
@@ -7,8 +9,12 @@ module Mini.Data.Set (
 
   -- * Construction
   empty,
-  fromList,
   singleton,
+  fromList,
+  fromAscList,
+  fromDescList,
+  fromDistinctAscList,
+  fromDistinctDescList,
 
   -- * Combination
   difference,
@@ -55,11 +61,16 @@ import Control.Applicative (
   (<|>),
  )
 import Data.Bifunctor (
+  bimap,
   first,
+  second,
  )
 import Data.Bool (
   bool,
  )
+import Data.Function (
+  on,
+ )
 import Prelude (
   Bool (
     False,
@@ -81,25 +92,32 @@ import Prelude (
   ),
   Semigroup,
   Show,
+  all,
   any,
   compare,
+  div,
   error,
   flip,
   foldl,
   foldr,
+  length,
   max,
   maybe,
   mempty,
   not,
   show,
+  splitAt,
   uncurry,
+  until,
   ($),
   (&&),
+  (*),
   (+),
   (-),
   (.),
   (<),
   (<$>),
+  (<*>),
   (<>),
   (==),
   (>),
@@ -121,10 +139,10 @@ data Set a
     R (Set a) a (Set a)
 
 instance (Eq a) => Eq (Set a) where
-  t1 == t2 = toAscList t1 == toAscList t2
+  (==) = (==) `on` toAscList
 
 instance (Ord a) => Ord (Set a) where
-  compare t1 t2 = compare (toAscList t1) (toAscList t2)
+  compare = compare `on` toAscList
 
 instance (Show a) => Show (Set a) where
   show = show . toAscList
@@ -170,13 +188,47 @@ set e f g h (R l a r) = h l a r (set e f g h l) (set e f g h r)
 empty :: Set a
 empty = E
 
+-- | /O(1)/ Make a set with a single element
+singleton :: a -> Set a
+singleton a = B E a E
+
 -- | /O(n log n)/ Make a set from a list of elements
 fromList :: (Ord a) => [a] -> Set a
 fromList = foldl (flip insert) empty
 
--- | /O(1)/ Make a set with a single element
-singleton :: a -> Set a
-singleton a = B E a E
+-- | /O(n)/ Make a set from a sorted list of elements
+fromAscList :: (Eq a) => [a] -> Set a
+fromAscList = fromDistinctAscList . essence
+
+-- | /O(n)/ Make a set from a sorted list of elements
+fromDescList :: (Eq a) => [a] -> Set a
+fromDescList = fromDistinctDescList . essence
+
+-- | /O(n)/ Make a set from a sorted list of distinct elements
+fromDistinctAscList :: [a] -> Set a
+fromDistinctAscList = go <*> power
+ where
+  go [] _ = E
+  go [a] _ = B E a E
+  go as n =
+    let len = length as
+        n' = n `div` 2
+        c = bool B L $ len == n
+        (l, a : r) = splitAt (len `div` 2) as
+     in c (go l n') a (go r n')
+
+-- | /O(n)/ Make a set from a sorted list of distinct elements
+fromDistinctDescList :: [a] -> Set a
+fromDistinctDescList = go <*> power
+ where
+  go [] _ = E
+  go [a] _ = B E a E
+  go as n =
+    let len = length as
+        n' = n `div` 2
+        c = bool B R $ len == n
+        (l, a : r) = splitAt (len `div` 2) as
+     in c (go r n') a (go l n')
 
 {-
  - Combination
@@ -184,19 +236,26 @@ singleton a = B E a E
 
 -- | /O(m log n)/ Subtract a set by another
 difference :: (Ord a) => Set a -> Set a -> Set a
-difference = foldr delete
+difference t1 t2 = set empty go go go t1
+ where
+  go _ _ _ _ _ = foldr delete t1 t2
 
 -- | /O(m log n)/ Intersect a set with another
 intersection :: (Ord a) => Set a -> Set a -> Set a
-intersection t1 t2 =
-  foldr
-    (\a b -> bool b (insert a b) (a `member` t2))
-    empty
-    t1
+intersection t1 t2 = set empty go go go t2
+ where
+  go _ _ _ _ _ =
+    fromDistinctAscList $
+      foldr
+        (\a b -> bool b (a : b) $ a `member` t2)
+        []
+        t1
 
 -- | /O(m log n)/ Unite a set with another
 union :: (Ord a) => Set a -> Set a -> Set a
-union = flip $ foldr insert
+union t1 t2 = set t2 go go go t1
+ where
+  go _ _ _ _ _ = foldr insert t1 t2
 
 -- | Unite a collection of sets
 unions :: (Foldable t, Ord a) => t (Set a) -> Set a
@@ -220,29 +279,214 @@ toDescList = foldl (flip (:)) []
 
 -- | /O(log n)/ Delete an element from a set
 delete :: (Ord a) => a -> Set a -> Set a
-delete a0 t = bool t (go t) (a0 `member` t)
+delete a t = bool t (delete' a t) $ a `member` t
+
+-- | /O(log n)/ Delete the maximum element from a set
+deleteMax :: (Ord a) => Set a -> Set a
+deleteMax t = maybe t (`delete'` t) $ lookupMax t
+
+-- | /O(log n)/ Delete the minimum element from a set
+deleteMin :: (Ord a) => Set a -> Set a
+deleteMin t = maybe t (`delete'` t) $ lookupMin t
+
+-- | /O(n)/ Keep the elements satisfying a predicate
+filter :: (a -> Bool) -> Set a -> Set a
+filter p = fromDistinctAscList . foldr (\a b -> bool b (a : b) $ p a) []
+
+-- | /O(log n)/ Insert an element into a set
+insert :: (Ord a) => a -> Set a -> Set a
+insert a t = bool (insert' a t) t $ a `member` t
+
+{-
+ - Partition
+ -}
+
+-- | /O(n)/ Partition a set with a predicate into @(true, false)@ subsets
+partition :: (a -> Bool) -> Set a -> (Set a, Set a)
+partition p =
+  bimap fromDistinctAscList fromDistinctAscList
+    . foldr
+      (\a -> bool second first (p a) (a :))
+      ([], [])
+
+-- | /O(n)/ Split a set by an element into @(lt, eq, gt)@ subsets
+split :: (Ord a) => a -> Set a -> (Set a, Bool, Set a)
+split a0 =
+  (\(lt, a, gt) -> (fromDistinctAscList lt, a, fromDistinctAscList gt))
+    . foldr
+      ( \a (lt, a', gt) -> case compare a a0 of
+          LT -> (a : lt, a', gt)
+          EQ -> (lt, True, gt)
+          GT -> (lt, a', a : gt)
+      )
+      ([], False, [])
+
+-- | /O(log n)/ Split a set by its maximum element
+splitMax :: (Ord a) => Set a -> Maybe (a, Set a)
+splitMax t = ((,) <*> flip delete' t) <$> lookupMax t
+
+-- | /O(log n)/ Split a set by its minimum element
+splitMin :: (Ord a) => Set a -> Maybe (a, Set a)
+splitMin t = ((,) <*> flip delete' t) <$> lookupMin t
+
+{-
+ - Query
+ -}
+
+-- | /O(m log n)/ Check whether two sets have no elements in common
+disjoint :: (Ord a) => Set a -> Set a -> Bool
+disjoint t1 t2 = set True go go go t1
  where
-  go =
+  go _ _ _ _ _ = not $ any (`member` t1) t2
+
+-- | /O(n log m)/ Check whether the elements of a set exist in the other
+isSubsetOf :: (Ord a) => Set a -> Set a -> Bool
+isSubsetOf t1 t2 = set (null t1) go go go t2
+ where
+  go _ _ _ _ _ = all (`member` t2) t1
+
+-- | /O(log n)/ Fetch the least element greater than or equal to the given one
+lookupGE :: (Ord a) => a -> Set a -> Maybe a
+lookupGE a0 = set Nothing go go go
+ where
+  go _ a _ recl recr = case compare a a0 of
+    LT -> recr
+    EQ -> Just a
+    GT -> recl <|> Just a
+
+-- | /O(log n)/ Fetch the least element strictly greater than the given one
+lookupGT :: (Ord a) => a -> Set a -> Maybe a
+lookupGT a0 = set Nothing go go go
+ where
+  go _ a _ recl recr = case compare a a0 of
+    LT -> recr
+    EQ -> recr
+    GT -> recl <|> Just a
+
+-- | /O(log n)/ Fetch the greatest element less than or equal to the given one
+lookupLE :: (Ord a) => a -> Set a -> Maybe a
+lookupLE a0 = set Nothing go go go
+ where
+  go _ a _ recl recr = case compare a a0 of
+    LT -> recr <|> Just a
+    EQ -> Just a
+    GT -> recl
+
+-- | /O(log n)/ Fetch the greatest element strictly less than the given one
+lookupLT :: (Ord a) => a -> Set a -> Maybe a
+lookupLT a0 = set Nothing go go go
+ where
+  go _ a _ recl recr = case compare a a0 of
+    LT -> recr <|> Just a
+    EQ -> recl
+    GT -> recl
+
+-- | /O(log n)/ Fetch the maximum element, or 'Nothing' if the set is empty
+lookupMax :: Set a -> Maybe a
+lookupMax = set Nothing go go go
+ where
+  go _ a r _ recr = set (Just a) go' go' go' r
+   where
+    go' _ _ _ _ _ = recr
+
+-- | /O(log n)/ Fetch the minimum element, or 'Nothing' if the set is empty
+lookupMin :: Set a -> Maybe a
+lookupMin = set Nothing go go go
+ where
+  go l a _ recl _ = set (Just a) go' go' go' l
+   where
+    go' _ _ _ _ _ = recl
+
+-- | /O(log n)/ Check whether an element is in a set
+member :: (Ord a) => a -> Set a -> Bool
+member a0 = set False go go go
+ where
+  go _ a _ recl recr = case compare a0 a of
+    LT -> recl
+    EQ -> True
+    GT -> recr
+
+-- | /O(1)/ Check whether a set is empty
+null :: Set a -> Bool
+null = set True go go go where go _ _ _ _ _ = False
+
+-- | /O(n)/ Get the size of a set
+size :: Set a -> Int
+size = set 0 go go go where go _ _ _ recl recr = 1 + recl + recr
+
+{-
+ - Validation
+ -}
+
+-- | /O(n^2)/ Check whether a set is internally height-balanced and ordered
+valid :: (Ord a) => Set a -> Bool
+valid = liftA2 (&&) balanced ordered
+ where
+  balanced =
     set
-      (error "Set.delete: L0")
-      ( \l a r _ _ ->
-          case compare a0 a of
-            LT -> deleteLl l a r
-            EQ -> substituteL l r
-            GT -> deleteLr l a r
-      )
-      ( \l a r _ _ ->
-          case compare a0 a of
-            LT -> deleteBl l a r
-            EQ -> substituteBr l r
-            GT -> deleteBr l a r
-      )
-      ( \l a r _ _ ->
-          case compare a0 a of
-            LT -> deleteRl l a r
-            EQ -> substituteR l r
-            GT -> deleteRr l a r
-      )
+      True
+      (\l _ r recl recr -> levels l - levels r == 1 && recl && recr)
+      (\l _ r recl recr -> levels l - levels r == 0 && recl && recr)
+      (\l _ r recl recr -> levels r - levels l == 1 && recl && recr)
+  levels = set 0 go go go
+   where
+    go _ _ _ recl recr = 1 + max recl recr :: Int
+  ordered = set True go go go
+   where
+    go l a r recl recr =
+      set True lt lt lt l
+        && set True gt gt gt r
+     where
+      lt _ la _ _ _ = la < a && recl && recr
+      gt _ ra _ _ _ = ra > a && recl && recr
+
+{-
+ - Helpers
+ -}
+
+-- O(n) 'nub' for sorted lists
+essence :: (Eq a) => [a] -> [a]
+essence [] = []
+essence (a : as) = essence' a as
+ where
+  essence' x1 [] = [x1]
+  essence' x1 (x2 : xs) =
+    let rest = essence' x2 xs
+     in bool (x1 : rest) rest $ x1 == x2
+
+-- The greatest power of 2 <= the length of a non-empty collection
+power :: (Foldable t) => t a -> Int
+power as = until (> length as) (* 2) 2 `div` 2
+
+{-
+ - Let this comment serve as your warning. Return from whence you came and your
+ - sanity will be spared. You have been admonished.
+ -}
+
+-- O(log n) Delete an element from a set without checking for membership
+delete' :: (Ord a) => a -> Set a -> Set a
+delete' a0 =
+  set
+    (error "Set.delete: L0")
+    ( \l a r _ _ ->
+        case compare a0 a of
+          LT -> deleteLl l a r
+          EQ -> substituteL l r
+          GT -> deleteLr l a r
+    )
+    ( \l a r _ _ ->
+        case compare a0 a of
+          LT -> deleteBl l a r
+          EQ -> substituteBr l r
+          GT -> deleteBr l a r
+    )
+    ( \l a r _ _ ->
+        case compare a0 a of
+          LT -> deleteRl l a r
+          EQ -> substituteR l r
+          GT -> deleteRr l a r
+    )
+ where
   deleteRl l a r =
     set
       (error "Set.delete: L1")
@@ -635,21 +879,9 @@ delete a0 t = bool t (go t) (a0 `member` t)
   popRightBR l a rl ra rr = first (checkRightB l a) $ popRightR rl ra rr
   popRightBL l a rl ra rr = first (checkRightB l a) $ popRightL rl ra rr
 
--- | /O(log n)/ Delete the maximum element from a set
-deleteMax :: (Ord a) => Set a -> Set a
-deleteMax t = maybe t (`delete` t) $ lookupMax t
-
--- | /O(log n)/ Delete the minimum element from a set
-deleteMin :: (Ord a) => Set a -> Set a
-deleteMin t = maybe t (`delete` t) $ lookupMin t
-
--- | /O(n log n)/ Keep the elements satisfying a predicate
-filter :: (Ord a) => (a -> Bool) -> Set a -> Set a
-filter p = foldr (\a b -> bool b (insert a b) $ p a) empty
-
--- | /O(log n)/ Insert an element into a set
-insert :: (Ord a) => a -> Set a -> Set a
-insert a0 =
+-- O(log n) Insert an element into a set without checking for membership
+insert' :: (Ord a) => a -> Set a -> Set a
+insert' a0 =
   set
     (B E a0 E)
     (\l a r _ _ -> insertL l a r)
@@ -811,140 +1043,3 @@ insert a0 =
       )
       (\lrl lra lrr _ _ -> L (B ll la (insertR lrl lra lrr)) a r)
       lr
-
-{-
- - Partition
- -}
-
--- | /O(n log n)/ Partition a set with a predicate into @(true, false)@ subsets
-partition :: (Ord a) => (a -> Bool) -> Set a -> (Set a, Set a)
-partition p =
-  foldr
-    (\a (t1, t2) -> bool (t1, insert a t2) (insert a t1, t2) $ p a)
-    (empty, empty)
-
--- | /O(n log n)/ Split a set by an element into @(lt, eq, gt)@ subsets
-split :: (Ord a) => a -> Set a -> (Set a, Bool, Set a)
-split a0 =
-  foldr
-    ( \a (t1, a', t2) -> case compare a a0 of
-        LT -> (insert a t1, a', t2)
-        EQ -> (t1, True, t2)
-        GT -> (t1, a', insert a t2)
-    )
-    (empty, False, empty)
-
--- | /O(log n)/ Split a set by its maximum element
-splitMax :: (Ord a) => Set a -> Maybe (a, Set a)
-splitMax t = (,deleteMax t) <$> lookupMax t
-
--- | /O(log n)/ Split a set by its minimum element
-splitMin :: (Ord a) => Set a -> Maybe (a, Set a)
-splitMin t = (,deleteMin t) <$> lookupMin t
-
-{-
- - Query
- -}
-
--- | /O(m log n)/ Check whether two sets have no elements in common
-disjoint :: (Ord a) => Set a -> Set a -> Bool
-disjoint t1 = not . any (`member` t1)
-
--- | /O(n log m)/ Check whether the elements of a set exist in the other
-isSubsetOf :: (Ord a) => Set a -> Set a -> Bool
-isSubsetOf t1 t2 = foldr (\a b -> a `member` t2 && b) True t1
-
--- | /O(log n)/ Fetch the least element greater than or equal to the given one
-lookupGE :: (Ord a) => a -> Set a -> Maybe a
-lookupGE a0 = set Nothing go go go
- where
-  go _ a _ recl recr = case compare a a0 of
-    LT -> recr
-    EQ -> Just a
-    GT -> recl <|> Just a
-
--- | /O(log n)/ Fetch the least element strictly greater than the given one
-lookupGT :: (Ord a) => a -> Set a -> Maybe a
-lookupGT a0 = set Nothing go go go
- where
-  go _ a _ recl recr = case compare a a0 of
-    LT -> recr
-    EQ -> recr
-    GT -> recl <|> Just a
-
--- | /O(log n)/ Fetch the greatest element less than or equal to the given one
-lookupLE :: (Ord a) => a -> Set a -> Maybe a
-lookupLE a0 = set Nothing go go go
- where
-  go _ a _ recl recr = case compare a a0 of
-    LT -> recr <|> Just a
-    EQ -> Just a
-    GT -> recl
-
--- | /O(log n)/ Fetch the greatest element strictly less than the given one
-lookupLT :: (Ord a) => a -> Set a -> Maybe a
-lookupLT a0 = set Nothing go go go
- where
-  go _ a _ recl recr = case compare a a0 of
-    LT -> recr <|> Just a
-    EQ -> recl
-    GT -> recl
-
--- | /O(log n)/ Fetch the maximum element, or 'Nothing' if the set is empty
-lookupMax :: Set a -> Maybe a
-lookupMax = set Nothing go go go
- where
-  go _ a r _ recr = set (Just a) go' go' go' r
-   where
-    go' _ _ _ _ _ = recr
-
--- | /O(log n)/ Fetch the minimum element, or 'Nothing' if the set is empty
-lookupMin :: Set a -> Maybe a
-lookupMin = set Nothing go go go
- where
-  go l a _ recl _ = set (Just a) go' go' go' l
-   where
-    go' _ _ _ _ _ = recl
-
--- | /O(log n)/ Check whether an element is in a set
-member :: (Ord a) => a -> Set a -> Bool
-member a0 = set False go go go
- where
-  go _ a _ recl recr = case compare a0 a of
-    LT -> recl
-    EQ -> True
-    GT -> recr
-
--- | /O(1)/ Check whether a set is empty
-null :: Set a -> Bool
-null = set True go go go where go _ _ _ _ _ = False
-
--- | /O(n)/ Get the size of a set
-size :: Set a -> Int
-size = set 0 go go go where go _ _ _ recl recr = 1 + recl + recr
-
-{-
- - Validation
- -}
-
--- | /O(n)/ Check whether a set is internally height-balanced and ordered
-valid :: (Ord a) => Set a -> Bool
-valid = liftA2 (&&) balanced ordered
- where
-  balanced =
-    set
-      True
-      (\l _ r recl recr -> levels l - levels r == 1 && recl && recr)
-      (\l _ r recl recr -> levels l - levels r == 0 && recl && recr)
-      (\l _ r recl recr -> levels r - levels l == 1 && recl && recr)
-  levels = set 0 go go go
-   where
-    go _ _ _ recl recr = 1 + max recl recr :: Int
-  ordered = set True go go go
-   where
-    go l a r recl recr =
-      set True lt lt lt l
-        && set True gt gt gt r
-     where
-      lt _ la _ _ _ = la < a && recl && recr
-      gt _ ra _ _ _ = ra > a && recl && recr
